#ifndef Controller_H
#define Controller_H

#include <iostream>
#include <sstream>
#include <list>
#include <string>

#include <boost/regex.hpp>

#include <QStringListModel>

#include <ExtendedTask.hpp>
#include <State.hpp>
#include <Client.hpp>
#include <MsgStructure.hpp>

#include <simple_logger.hpp>

namespace hatptester{

class HATPController: public QObject{
	Q_OBJECT

private:
	//To remember previous state
	std::string entitiesMessage;
	std::string methodsMessage;
	std::string actionsMessage;

	HATPState state;
	std::map<std::string, ExtendedTask> extendedTasks;
	ExtendedTask* currentTask;
	
	QStringListModel tasksModel;
	
	static std::string address; //Value defined in View.cpp
	static unsigned int port;
	HATPClient client;
	bool correctlyInitialised;
	
	int requestId;
	
	#define METHODS 0
	#define ACTIONS 1	
	//Tasktype describes the type of task to add to the state : either 0 for Methods or 1 for Actions
	bool hasSameTasks(bool taskType){
		std::string askTasks;
		if(taskType==METHODS){
			askTasks="{\"HATP-REQ\": { \"REQ-TYPE\": \"getTasks\"}}";
			LOG(trace)<<"Will request the method list";
		}else{
			askTasks="{\"HATP-REQ\": { \"REQ-TYPE\": \"getActions\"}}";
			LOG(trace)<<"Will request the actions list";
		}
		std::pair<bool, std::string> tasks=client.requestValue(askTasks);
		if(tasks.first){
			if(taskType==METHODS){
				if(tasks.second==methodsMessage){ //The entities did not change since last time
					return true;
				}else{
					methodsMessage=tasks.second;
					return false;
				}
			}else{
				if(tasks.second==actionsMessage){ //The entities did not change since last time
					return true;
				}else{
					actionsMessage=tasks.second;
					return false;
				}
			}
		}
		LOG(error)<<"Could get an answer form HATP";
		methodsMessage="";
		actionsMessage="";
		return false;
	}
	
	bool initialiseTasks(bool taskType){
		boost::smatch result;
		boost::regex taskAnswerRegex;
		std::string tasksMessage;
		if(taskType==METHODS){
			if(methodsMessage=="") return false;
			tasksMessage=methodsMessage;
			taskAnswerRegex=methodAnswerRegex;
		}else{
			if(actionsMessage=="") return false;
			tasksMessage=actionsMessage;
			taskAnswerRegex=actionAnswerRegex;
		}
		if(boost::regex_match(tasksMessage, result, taskAnswerRegex)){
			std::string tasks=result[1];
			boost::regex taskRegex;
			if(taskType==METHODS){
				taskRegex=methodRegex;
			}else{
				taskRegex=actionRegex;
			}
			
			boost::sregex_token_iterator taskIterator(tasks.begin(), tasks.end(), taskRegex), endTask;

			while(taskIterator != endTask){
				boost::smatch resultTask;
				const std::string taskToProcess=*taskIterator;
				
				if(boost::regex_match(taskToProcess, resultTask, taskRegex)){
					std::string parameters=resultTask[2];
					boost::sregex_token_iterator parameterIterator(parameters.begin(), parameters.end(), wordRegex), endParameter;
					Task::Parameters temp;
					unsigned long int i=0;
					while(parameterIterator != endParameter){
						std::stringstream tempName; //TODO replace by actual name from domain
						tempName<<*parameterIterator<<i;
						temp.push_back(Task::Parameter(*parameterIterator, tempName.str()));
						++parameterIterator;
						++i;
					}
	
					if(taskType==METHODS){
						state.addMethod(Method(resultTask[1], temp));
					}else{
						state.addAction(Action(resultTask[1], temp));
					}
				}
				++taskIterator;
			}
			return true;
		}else{
			LOG(warning)<<"Answer from HATP is not correctly interpreted";
		}
		
		LOG(error)<<"Error while requesting the tasks (methods and actions), return error flag";
		return false;
	}
	
	bool hasSameEntities(){
		std::string askEntities="{\"HATP-REQ\": { \"REQ-TYPE\": \"getEntities\"}}";
		LOG(trace)<<"Request the list of entities";
		std::pair<bool, std::string> entities=client.requestValue(askEntities);
		if(entities.first){
			if(entities.second==entitiesMessage){ //The entities did not change since last time
				return true;
			}else{
				entitiesMessage=entities.second;
				return false;
			}
		}
		LOG(error)<<"Could not get an answer from HATP";
		entitiesMessage="";
		return false;
	}
	
	bool initialiseEntities(){
		bool success=false;
		
		boost::smatch result;
		if(entitiesMessage=="") return false;
		if(boost::regex_match(entitiesMessage, result, entityAnswerRegex)){
			success=true;
			std::string entities=result[1];
							
			boost::sregex_token_iterator entitiesIterator(entities.begin(), entities.end(), entityRegex), endEntities;

			while(entitiesIterator != endEntities){
				boost::smatch resultEntity;
				const std::string entityToProcess=*entitiesIterator;
				if(boost::regex_match(entityToProcess, resultEntity, entityRegex)){
					if(resultEntity.size()!=3){
						success=false;
					}
					state.addEntity(Entity(resultEntity[2], resultEntity[1]));
					success=success and true;
				}
				++entitiesIterator;
			}
		}else{
			LOG(warning)<<"Answer from HATP is not correctly interpreted";
		}
		
		return success;
	}
	
	void initialiseTaskModel(){
		QStringList tasks;
		//TODO Remove those marking and put unselectable labels instead (with different backgroung color)
		tasks << " -- Methods : ";
		
		for(std::map<std::string, Method>::const_iterator it=state.getMethods().begin(); it!=state.getMethods().end(); ++it){
			tasks << it->first.c_str();
			extendedTasks.insert(std::make_pair(it->first, ExtendedTask(state, it->first)));
		}
		tasks << "";

		tasks << " -- Actions : ";
		for(std::map<std::string, Action>::const_iterator it=state.getActions().begin(); it!=state.getActions().end(); ++it){
			tasks << it->first.c_str();
			extendedTasks.insert(std::make_pair(it->first, ExtendedTask(state, it->first)));
		}

		tasksModel.setStringList(tasks);
	}
	
	void clearTaskModel(){
		QStringList tasks;
		tasksModel.setStringList(tasks);
	}
	
	bool initialiseModel(){
		bool success=false;
		correctlyInitialised=false;
		bool changed=false;
		
		if(client.isConnected()){
			//Initialise state using requests to HATP
			if(!hasSameEntities()){
				state.clearEntities();
				success=initialiseEntities();
				changed=true;
			} else success=true;

			if(success){
				if(!hasSameTasks(METHODS)){
					state.clearMethods();
					success=initialiseTasks(METHODS); //Get Methods
					changed=true;
				} else success=true;
				
				if(success){				
					if(!hasSameTasks(ACTIONS)){
						state.clearActions();
						success=initialiseTasks(ACTIONS); //Get Actions
						changed=true;
					} else success=true;
					
					if(success){		
						if(changed){//Initialise the different models for Qt GUI
							initialiseTaskModel();
							
							currentTask=NULL;
						}
						requestId=0;
						correctlyInitialised=true;
						LOG(info)<<"The structure representing the domain has been succesfully initialised";
					}else{
						LOG(error)<<"Could not get the actions";
					}
				}else{
					LOG(error)<<"Could not get the methods";
				}
			}else{
				LOG(error)<<"Could not get the entities";
			}
		}else{
			LOG(error)<<"Client failed to connect";
			return false;
		}
		
		if(!success){
			LOG(warning)<<"Failed to initialise from HATP, please check that HATP is running correctly";
			state.clearEntities();
			state.clearMethods();
			state.clearActions();
			clearTaskModel();
			entitiesMessage="";
			methodsMessage="";
			actionsMessage="";
		}
		if(changed){
			emit sendStateChanged();
		}
		return success;
	}
	
	HATPController():client("hatptester", address, port), correctlyInitialised(false){
		THREAD_NAME("Controller");
		
		initialiseModel();
	}
	HATPController(HATPController const&);
	HATPController& operator=(HATPController const&);
	
	void sendRequest(bool stopAtFirstPlan){
		if(currentTask==NULL) return;
		
		//True means it stops after the first valid plan found
		//False means it computes all possible plans
		std::stringstream request;
		
		//Header
		request<<"{\"HATP-REQ\": { \"REQ-TYPE\": \""<< (stopAtFirstPlan?"firstPlan":"plan") <<"\", \"REQ-ID\": "<<requestId<<", ";
		++requestId;
		
		//Task
		request<<"\"TASK-NAME\": \""<<currentTask->getName()<<"\", ";
		
		//Parameters
		request<<"\"PARAMETERS\": [";
		std::pair<bool, const Task*> result;
		result=state.getTask(currentTask->getName());
		if(!result.first){
			return;
		}
		const Task::Parameters& parameters=result.second->getParameters();
		size_t index=0;
		for(Task::Parameters::const_iterator it=parameters.begin(); it!=parameters.end(); ++it){
			++index;
			request<<"\""<<it->value<<"\"";
			if(index!=parameters.size()){
				request<<", ";
			}			
		}
		
		//Footer
		request<<"]}}";
		
		LOG(trace)<<"Send the request : "<<request.str();
		
		//Send the request
		if(!client.sendMessage(request.str())){
			LOG(error)<<"Error while sending the message";
			this->exit();
		}
	}
	
public:	
	static void setInstance(const std::string& serverAddress, unsigned int serverPort){
		address=serverAddress;
		port=serverPort;
	}

	static HATPController& getInstance(){
		static HATPController instance;
		return instance;
	}
	
	~HATPController(){
		client.stopThread();
	}
	
	bool isConnected(){
		return(client.isConnected());
	}
	
	bool ishatpConnected(){
		return(this->isConnected() && client.ishatpConnected());
	}
	
	bool isReady(){
		return(this->ishatpConnected()&&correctlyInitialised);
	}
	
	void exit(){
		client.stopThread();
		emit sendStop();
	}
	
	void reloadModel(){
		initialiseModel();
	}
	
	const std::map<std::string, Entity> getEntities() const{
		return(state.getEntities());
	}
	
	void sendFirstPlanRequest(){
		sendRequest(true); //True means it stops after the first valid plan found
	}
	
	void sendAllPlansRequest(){
		sendRequest(false); //False means it computes all possible plans
	}
	
	void sendMaxTime(int value){ //value<=0 means to send an unlimited planning time, any other value means limited planning time
		std::stringstream request;
		request<<"{\"HATP-REQ\": { \"REQ-TYPE\": ";
		
		if(value<=0){
			request<<"\"unsetMaxTime\"";
		}else{
			request<<"\"setMaxTime\", \"MAXTIME\": "<<value;
		}
		
		request<<"}}";
		
		LOG(trace)<<"Send the request : "<<request.str();
		
		//Send the request
		if(!client.sendMessage(request.str())){
			LOG(error)<<"Error while sending the message";
			this->exit();
		}
	}
	
	QStringListModel& getTasks(){
		return tasksModel;
	}
	
	QStandardItemModel* getTaskParameter(const std::string& task){
		std::map<std::string, ExtendedTask>::iterator result;
		result=extendedTasks.find(task);
		if(result!=extendedTasks.end()){
			currentTask=&result->second;
			return result->second.getParametersModel();
		}
		return NULL;
	}
	
	bool isReadyTask(const std::string& task){
		std::pair<bool, const Task*> result;
		result=state.getTask(task);
		if(result.first){
			return result.second->isReady();
		}
		return false;
	}
	
	bool isReadyCurrentTask(){
		if(currentTask==NULL) return false;
		
		std::pair<bool, const Task*> result;
		result=state.getTask(currentTask->getName());
		if(result.first){
			return result.second->isReady();
		}
		return false;
	}
	
	const std::string getCurrentTaskName() const{
		if(currentTask==NULL) return "";
		
		return currentTask->getName();
	}
	
	std::set<std::string> getMissingParameters(const std::string& task){
		std::set<std::string> missingParameters;
	
		std::pair<bool, const Task*> result;
		result=state.getTask(task);
		if(result.first && !result.second->isReady()){
			const Task::Parameters& parameters=result.second->getParameters();
			for(Task::Parameters::const_iterator it=parameters.begin(); it!=parameters.end(); ++it){
				if(it->value==""){
					missingParameters.insert(it->name+" ["+it->type+"]");
				}
			}
		}
		return missingParameters;
	}
	
	const std::string getLongestTaskName() const{
		std::string longest="";
		for(HATPState::MethodsMap::const_iterator it=state.getMethods().begin(); it!=state.getMethods().end(); ++it){
			if(it->first.length()>longest.length()){
				longest=it->first;
			}
		}
		for(HATPState::ActionsMap::const_iterator it=state.getActions().begin(); it!=state.getActions().end(); ++it){
			if(it->first.length()>longest.length()){
				longest=it->first;
			}
		}
		return longest;
	}
	
signals:
	void sendStop();
	void sendStateChanged();
};

}; //namespace

#endif //Controller_H;
